This patch implements the data plane of reading data from bootstrap blob
file over fscache for inline layout.

For the heading non-inline part, the data plane for non-inline layout is
resued, while only the tail packing part needs special handling.

Signed-off-by: Jeffle Xu <jeffl...@linux.alibaba.com>
---
 fs/erofs/fscache.c | 43 +++++++++++++++++++++++++++++++++++++++++--
 1 file changed, 41 insertions(+), 2 deletions(-)

diff --git a/fs/erofs/fscache.c b/fs/erofs/fscache.c
index df56562f33c4..254b3e72ab4d 100644
--- a/fs/erofs/fscache.c
+++ b/fs/erofs/fscache.c
@@ -85,8 +85,9 @@ static int erofs_fscache_readpage_noinline(struct page *page,
 {
        struct fscache_cookie *cookie = fsmap->m_ctx->cookie;
        /*
         * 1) For FLAT_PLAIN layout, the output map.m_la shall be equal to o_la,
         * and the output map.m_pa is exactly the physical address of o_la.
         * 2) For CHUNK_BASED layout, the output map.m_la is rounded down to the
         * nearest chunk boundary, and the output map.m_pa is actually the
         * physical address of this chunk boundary. So we need to recalculate
@@ -97,6 +98,40 @@ static int erofs_fscache_readpage_noinline(struct page *page,
        return erofs_fscache_read_page(cookie, page, start);
 }
 
+static int erofs_fscache_readpage_inline(struct page *page,
+                                        struct erofs_fscache_map *fsmap)
+{
+       struct inode *inode = page->mapping->host;
+       struct super_block *sb = inode->i_sb;
+       struct erofs_buf buf = __EROFS_BUF_INITIALIZER;
+       erofs_blk_t blknr;
+       size_t offset, len;
+       void *src, *dst;
+
+       /*
+        * For inline (tail packing) layout, the offset may be non-zero, which
+        * can be calculated from corresponding physical address directly.
+        * Currently only flat layout supports inline (FLAT_INLINE), and the
+        * output map.m_pa is exactly the physical address of o_la in this case.
+        */
+       offset = erofs_blkoff(fsmap->m_pa);
+       blknr = erofs_blknr(fsmap->m_pa);
+       len = fsmap->m_llen;
+
+       src = erofs_read_metabuf(&buf, sb, blknr, EROFS_KMAP);
+       if (IS_ERR(src))
+               return PTR_ERR(src);
+
+       dst = kmap(page);
+       memcpy(dst, src + offset, len);
+       memset(dst + len, 0, PAGE_SIZE - len);
+       kunmap(page);
+
+       erofs_put_metabuf(&buf);
+
+       return 0;
+}
+
 static int erofs_fscache_do_readpage(struct page *page)
 {
        struct inode *inode = page->mapping->host;
@@ -126,8 +161,12 @@ static int erofs_fscache_do_readpage(struct page *page)
        if (ret)
                return ret;
 
+       if (map.m_flags & EROFS_MAP_META)
+               return erofs_fscache_readpage_inline(page, &fsmap);
+
        switch (vi->datalayout) {
        case EROFS_INODE_FLAT_PLAIN:
+       case EROFS_INODE_FLAT_INLINE:
        case EROFS_INODE_CHUNK_BASED:
                return erofs_fscache_readpage_noinline(page, &fsmap);
        default:
@@ -85,8 +85,9 @@ static int erofs_fscache_readpage_noinline(struct page *page,
        struct fscache_cookie *cookie = fsmap->m_ctx->cookie;
        /*
-        * 1) For FLAT_PLAIN layout, the output map.m_la shall be equal to o_la,
-        * and the output map.m_pa is exactly the physical address of o_la.
+        * 1) For FLAT_PLAIN and FLAT_INLINE (the heading non tail packing part)
+        * layout, the output map.m_la shall be equal to o_la, and the output
+        * map.m_pa is exactly the physical address of o_la.
         * 2) For CHUNK_BASED layout, the output map.m_la is rounded down to the
         * nearest chunk boundary, and the output map.m_pa is actually the
-- 
2.27.0

--
Linux-cachefs mailing list
Linux-cachefs@redhat.com
https://listman.redhat.com/mailman/listinfo/linux-cachefs

Reply via email to